require 'minitest'
# require 'minitest/human_reporter_plugin'

class Minitest::Test
  def assert_equal expected, actual, msg = nil
    return true if expected == actual

    handler = handler(expected, actual)

    return super(expected, actual, msg) if expected.nil? || actual.nil? || handler.nil?

    raise Minitest::Assertion, handler.message()
  end

  def handler(expected, actual)
    case expected
    when Array
      ArrayHandler.new(expected, actual)
    when Hash
      HashHandler.new(expected, actual)
    when String
      StringHandler.new(expected, actual)
    end
  end
end

class Handler
  RED = 31
  GREEN = 32

  attr_reader :expected, :actual

  def initialize(expected, actual)
    @expected = expected
    @actual = actual
  end

  def message
    raise "plz implement me"
  end

  def intersection
    raise "plz implement me"
  end

  def missing
    raise "plz implement me"
  end

  def remaining
    raise "plz implement me"
  end
end

class ArrayHandler < Handler
  def intersection
    @intersection ||= expected & actual
  end

  def missing
    @missing ||= expected - intersection
  end

  def remaining
    @remaining ||= actual - intersection
  end

  def message
    result = ""
    print_with_colour(RED, result, expected, "Expected", intersection)
    print_with_colour(GREEN, result, actual, "  Actual", intersection)

    if missing.any? || remaining.any?
      result << "\n"

      result << "   Missing: #{missing.inspect}\n" unless missing.empty?
      result << " Remaining: #{remaining.inspect}\n" unless remaining.empty?
    end

    if expected.to_set == actual.to_set
      result << "\n    Both arrays have the same elements, but order matters in array comparisons."
      result << " If element order does not matter, you can call '.to_set' on both before asserting.\n"
    end

    result
  end

  private

  def print_with_colour(colour, message, enumerable, title, intersection)
    message << "  #{title}: ["

    first = true
    enumerable.each do |e|
      message << ", " unless first
      first = false

      if !intersection.include?(e)
        message << "\e[#{colour}m#{e.inspect}\e[0m"
      else
        message << e.inspect
      end
    end

    message << "]\n"
  end
end

class HashHandler < Handler
  def intersection
    @intersection ||= expected.select do |key, value|
      actual.key?(key) && actual[key] == value
    end
  end

  def missing
    @missing ||= expected.reject do |key, value|
      intersection.key?(key) && intersection[key] == value
    end
  end

  def remaining
    @remaining ||= actual.reject do |key, value|
      intersection.key?(key) && intersection[key] == value
    end
  end

  def message
    result = ""
    print_with_colour(RED, result, expected, "Expected", intersection)
    print_with_colour(GREEN, result, actual, "  Actual", intersection)

    if missing.any? || remaining.any?
      result << "\n"

      result << "   Missing: #{missing.inspect}\n" unless missing.empty?
      result << " Remaining: #{remaining.inspect}\n" unless remaining.empty?
    end

    result
  end

  private

  def print_with_colour(colour, message, enumerable, title, intersection)
    message << "  #{title}: {"

    first = true
    enumerable.each do |k, v|
      message << ", " unless first
      first = false

      item_message = "#{k.inspect}=>#{v.inspect}"
      if intersection.key?(k) && intersection[k] == v
        message << item_message
      else
        message << "\e[#{colour}m#{item_message}\e[0m"
      end
    end

    message << "}\n"
  end
end

class StringHandler < Handler
  def message
    result = ""

    if actual.start_with?(expected)
      result << "  Expected: #{expected.inspect}\n"
      result << "    Actual: \"#{expected}\e[#{GREEN}m#{actual[expected.size..-1]}\e[0m\"\n"
    elsif expected.start_with?(actual)
      result << "  Expected: \"#{actual}\e[#{RED}m#{expected[actual.size..-1]}\e[0m\"\n"
      result << "    Actual: #{actual.inspect}\n"
    else
      intersection_index = 0

      1.upto(expected.size) do |index|
        string_prefix = expected[0...index]

        break unless actual.start_with?(string_prefix)

        intersection_index += 1
      end

      intersection = expected[0...intersection_index]

      result << "  Expected: \"#{intersection}\e[#{RED}m#{expected[intersection.size..expected.size-1]}\e[0m\"\n"
      result << "    Actual: \"#{intersection}\e[#{GREEN}m#{actual[intersection.size..actual.size-1]}\e[0m\"\n"
    end

    result
  end
end
